/* rsaencrypt v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>
 * 
 * Encrypts a file with a keyfile cretaed with rsakeygen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* includes */

/* ANSI C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* GMP library */
#include <gmp.h>

/* UNIX */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


/* defines */

#define VERSION "rsaencrypt v1.0 Copyright (C) 2003 tomasliq <tom@eggdrop.ch>"
#define NUMVER "010000"


/* prototypes */
int main(int argc, char *argv[]);
void usage(int argc, char *argv[]);
int file_exists(char *f);
void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n);
void err(int num);
void plain2crypt(FILE *fp, FILE *fc, unsigned char **e, unsigned char **n);

/* functions */
int main(int argc, char *argv[])
{
	FILE *key, *plain, *crypt;
	unsigned char *e, *n;
	char type;
	
	/* print version */
	printf("%s\n", VERSION);

	/* check number of arguments */
	if (argc < 4) { usage(argc, argv); exit(1); }
	
	/* check to see whether the files exist or not */
	if (!file_exists(argv[1]))
	{
		printf("Error: Keyfile does not exist\n");
		exit(1);
	}
	
	if (!file_exists(argv[2]))
	{
		printf("Error: Plainfile does not exist\n");
		exit(1);
	}

	if (file_exists(argv[3]))
	{
		printf("Error: Cryptfile already exists\n");
		exit(1);
	}
	
	/* open the files */
	if ((key = fopen(argv[1], "r")) == NULL)
	{
		printf("Error: Can't open keyfile %s for reading\n", argv[1]);
		printf("Have you the required permissions?\n");
		exit(1);
	}

	if ((plain = fopen(argv[2], "r")) == NULL)
	{
		printf("Error: Can't open plainfile %s for reading\n", argv[2]);
		printf("Have you the required permissions?\n");
		exit(1);
	}

	if ((crypt = fopen(argv[3], "w")) == NULL)
	{
		printf("Error: Can't open cryptfile %s for writing\n", argv[1]);
		printf("Have you the required permissions?\n");
		exit(1);
	}
	
	/* read the key */
	readkey(key, &type, &e, &n);

	/* read plainfile */
	plain2crypt(plain, crypt, &e, &n);

	/* close the keyfile */
	fclose(key);

	/* print the key */

	free(e);
	free(n);

	/* and exit */
	return 0;
}

/* prints the usage */
void usage(int argc, char *argv[])
{
        printf("Usage: %s <publickey> <plainfile> <cryptfile>\n", argv[0]);
}

/* returns 1 if a file exists or 0 if not*/
int file_exists(char *f) {
        struct stat s;
        return ((stat(f,&s)==0)?1:0);
}

void readkey(FILE *f, char *type, unsigned char **de, unsigned char **n)
{
	int c=0; /* stores the character read by fgetc */
	int i; /* used for loops */
	char numver[6];
	fpos_t filepos; /* stores the file position */
	long len; /* length of D/E or N */
	
	*type=0;

	/* the file starts with RSA */
	c = fgetc(f);
	if (c != 'R') err(1);
	c = fgetc(f);
	if (c != 'S') err(1);
	c = fgetc(f);
	if (c != 'A') err(1);

	/* read the numver */
	for (i=0;i<6;i++)
	{
		c = fgetc(f);
		if (c == EOF) err(1);
		if (!((c>='0')&&(c<='9'))) err(1);
		numver[i]=c;
	}
	if (!(numver[0] == '0' && numver[1] == '1')) err(3);

	/* read type of file */
	*type = fgetc(f);
	c = fgetc(f);
	if (c != ' ') err(1);
	if (*type != 'P') err(2);
	
	/* save the current position */
	if (fgetpos(f, &filepos) < 0)
	{
		printf("Error: fgetpos() failed\n");
		exit(1);
	}

	/* how long is D/E ? */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		if (c < 32) err(1);
		i++;
	}
	len = i;

	/* now allocate memory */
	*de=(unsigned char *)malloc(len+1);
	if (*de == NULL) { printf("Error: No memory available\n"); exit(1); }

	c = fsetpos(f, &filepos);
	if (c == -1)
	{
		printf("Error: fsetpos() failed\n");
		exit(1);
	}

	/* write D/E into memory */
	i=0;
	while((c = fgetc(f)) != 32)
	{
		*(*de+i)=c;
		i++;
	}
	*(*de+i)='\0';

        /* save the current position */
        if (fgetpos(f, &filepos) < 0)
        {
                printf("Error: fgetpos() failed\n");
                exit(1);
        }

        /* how long is N ? */
        i=0;
        while((c = fgetc(f)) != '\n') /* last char is \n */
        {
                if (c <= 32) err(1);
                i++;
        }
        len = i;

        /* now allocate memory */
        *n=(unsigned char *)malloc(len+1);
	if (*n == NULL) { printf("Error: No memory available\n"); exit(1); }

        c = fsetpos(f, &filepos);
        if (c == -1)
        {
                printf("Error: fsetpos() failed\n");
                exit(1);
        }

        /* write N into memory */
        i=0;
        while((c = fgetc(f)) != '\n')
        {
                *(*n+i)=c;
                i++;
        }
        *(*n+i)='\0';

	return;
}

void err(int num)
{
	switch(num)
	{
		case 1:
			printf("Error: This is not a valid rsakeygen file\n");
			break;
		case 2:
			printf("Error: This is not a public key file \n");
			break;
		case 3:
			printf("Error: Version should be 1.x.x\n");
			break;
	}
	exit(1);
}

void plain2crypt(FILE *fp, FILE *fc, unsigned char **e, unsigned char **n)
{
	int i=0,i2=0,x=0;
	unsigned char *tmpp, *hexp;
	mpz_t ne, nn, tmp1, tmp2, tmp3, base, encnum, crypt;
	int itmp1, itmp2;
	float procent;

	printf("Reading plainfile...");
	fflush(stdout);

	while (fgetc(fp) != EOF) i++;

	/* we don't need +1, because we don't store a \0 */
	tmpp = (unsigned char *)malloc(i);
	
	hexp = (unsigned char *)malloc(i*2+1);

	rewind(fp);
	
	for (x=0; x<i; x++)
	{
		tmpp[x]=fgetc(fp);
	}

	printf("OK\n");

	printf("Preparing plainfile.");
	fflush(stdout);

	i2=0;
	for(x=0; x<i; x++)
	{
		char c=tmpp[x];
		char hexnum;
		char hex;
		
		hexnum=0;
		if (c & 1) hexnum+=1;
		if (c & 2) hexnum+=2;
		if (c & 4) hexnum+=4;
		if (c & 8) hexnum+=8;
		if (hexnum <= 8) hex = '1' + hexnum;
		if (hexnum >= 9) hex = 'a' + (hexnum-9);
		
		hexp[i2]=hex;
		i2++;

		hexnum=0;
		if (c &  16) hexnum+=1;
		if (c &  32) hexnum+=2;
		if (c &  64) hexnum+=4;
		if (c & 128) hexnum+=8;
		if (hexnum <= 8) hex = '1' + hexnum;
		if (hexnum >= 9) hex = 'a' + (hexnum-9);

		hexp[i2]=hex;
		i2++;
		
		if (x%1048576 == 0) { printf("."); fflush(stdout); }
	}
	hexp[i2]='\0';
	printf("OK\n");

	printf("Encrypting %d bytes ", i);
	free(tmpp);

	mpz_init(ne); mpz_init(nn);
	mpz_init(tmp1); mpz_init(tmp2); mpz_init(tmp3);
	mpz_init(base);
	mpz_init(encnum); mpz_init(crypt);

	mpz_set_ui(base, 223);

	mpz_set_ui(ne, *(*e)-33);
	for (i=1;i<strlen((char *)*e);i++)
	{
		mpz_pow_ui(tmp1, base, i);
		mpz_mul_ui(tmp2, tmp1, *(*e+i)-33);
		mpz_set(tmp1, ne);
		mpz_add(ne, tmp1, tmp2);
	}

	
	mpz_set_ui(nn, *(*n)-33);
	for (i=1;i<strlen((char *)*n);i++)
	{
		mpz_pow_ui(tmp1, base, i);
		mpz_mul_ui(tmp2, tmp1, *(*n+i)-33);
		mpz_set(tmp1, nn);
		mpz_add(nn, tmp1, tmp2);
	}

	itmp1=strlen((char *)hexp);
	itmp2=strlen(mpz_get_str(NULL, 17, ne))-1;
	
	if (itmp2%2==1) itmp2--;
        printf("(blocksize is %d bytes). This can take a while:\n", itmp2);

	printf("\r%10d of %10d blocks encrypted %3d%%", 0, (itmp1+itmp2-1)/itmp2, 0);
	fflush(stdout);
	procent=(float)100.0/(itmp1/itmp2);

	fprintf(fc, "RSA%sC ", NUMVER);
	
	x=0;
	for (i=0;i<itmp1;i+=itmp2)
	{
		unsigned char *tmp;
		
		tmp = (unsigned char *)malloc(itmp2+1);
		strncpy((char *)tmp, (char *)&hexp[i], itmp2);
		tmp[itmp2]='\0';

		mpz_set_str(encnum, (char *)tmp, 17);
		mpz_powm(crypt, encnum, ne, nn);

		x++;
		printf("\r%10d of %10d blocks encrypted %3.0f%%", x, (itmp1+itmp2-1)/itmp2, (procent*x)>100?100:procent*x);
		fflush(stdout);
		
		mpz_set(tmp1,crypt);
		while(mpz_cmp_ui(tmp1,0) != 0)
		{
			mpz_mod(tmp2, tmp1, base);
			fputc(atoi(mpz_get_str(NULL,10,tmp2))+33, fc);
			mpz_sub(tmp3, tmp1, tmp2);
			mpz_tdiv_q(tmp1, tmp3, base);
		}
		fputc(' ', fc);
		
		/* printf("%s\n", mpz_get_str(NULL,10,crypt)); */
	}
	printf("\n");

	free(hexp);

	fputc('\n', fc);

	printf("Encrypting done!\n");

	return;
}
